import httplib
import cStringIO as StringIO
import struct

# A httplib-like connector for the AJP 1.3 protocol used by mod_jk2.
# http://jakarta.apache.org/tomcat/tomcat-4.1-doc/jk2/common/AJPv13.html

__all__ = ["AJK13Connection", "AJP13Response", "AJP13Error"]


##############################################################################
# Constants
##############################################################################

AJP_PORT = 8009

# Packet types
AJP_PACKET_FORWARD_REQUEST = 2
AJP_PACKET_SEND_BODY_CHUNK = 3
AJP_PACKET_SEND_HEADERS = 4
AJP_PACKET_END_RESPONSE = 5

# Methods
AJP_METHODS = [
    None, "OPTIONS", "GET", "HEAD", "POST", "PUT", "DELETE", "TRACE",
    "PROPFIND", "PROPPATCH", "MKCOL", "COPY", "MOVE", "LOCK", "UNLOCK",
    "ACL", "REPORT", "VERSION-CONTROL", "CHECKIN", "CHECKOUT",
    "UNCHECKOUT", "SEARCH", "MKWORKSPACE", "UPDATE", "LABEL", "MERGE",
    "BASELINE-CONTROL", "MKACTIVITY"]

# Headers
AJP_HEADERS = [
    None, "accept", "accept-charset", "accept-encoding", "accept-language",
    "authorization", "connection", "content-type", "content-length",
    "cookie", "cookie2", "host", "pragma", "referer", "user-agent"]

# Attributes
AJP_ATTRIBUTE_QUERY_STRING = 5


##############################################################################
# Packets
##############################################################################

class AJP13Error(Exception):
    """Something unexpected happened."""

# This allows us to distinguish between bytes, booleans and integers.
class Byte:
    def __init__(self, value):
        self.value = value

class AJPPacket:
    def __init__(self, data = ""):
        self.data = data

    # List-like methods for adding data to the packet

    def extend(self, items):
        for item in items:
            self.append(item)

    def append(self, item):
        if isinstance(item, Byte):
            self.data += struct.pack("B",  item.value)
        elif type(item) == type(0):
            self.data += struct.pack(">H",  item)
        elif type(item) == type(""):
            self.append(len(item))
            self.data += item + "\x00"
        else:
            raise ValueError, "can't encode %s" % repr(item) # type(item)

    # Methods for sending and receiving data over the network

    def transmit(self, sock):
        packet = AJPPacket()
        packet.extend((0x1234, self.data))
        packet = packet.data[:-1]
        sock.sendall(packet)

    def receive(self, fp):
        magic = fp.read(2)
        if magic != "AB":
            raise AJP13Error, "bad magic number %s" % repr(magic)
        [length] = struct.unpack(">H", fp.read(2))
        self.data = fp.read(length)

    # Methods for extracting data from the packet

    def __read(self, fmt):
        size = struct.calcsize(fmt)
        item, self.data = self.data[:size], self.data[size:]
        return struct.unpack(fmt, item)[0]

    def read_byte(self):
        return int(self.__read("B"))

    def read_int(self):
        return int(self.__read(">H"))

    def read_string(self):
        size = self.read_int()
        if size & 0xFFF0 == 0xA000:
            return AJP_HEADERS[size & 0x000F]
        else:
            ret = self.__read("%ds" % size)
            null = self.read_byte()
            if null:
                raise AJP13Error, "found 0x%02x after string" % null
            return ret


##############################################################################
# httplib-style classes
##############################################################################

class AJP13Message(httplib.HTTPMessage):

    def __init__(self, packet):
        self._packet = packet
        apply(httplib.HTTPMessage.__init__, (self, None, 0))
        del self._packet

    def readheaders(self):
        self.dict = {}
        self.unixfrom = []
        self.headers = []
        self.status = ""
        for header in xrange(self._packet.read_int()):
            name = self._packet.read_string()
            value = self._packet.read_string()
            self.headers.append("%s: %s" % (name, value))
            self.addheader(name.lower(), value)


class AJP13Response(httplib.HTTPResponse):

    def __init__(self, *args, **kwargs):
        apply(httplib.HTTPResponse.__init__, (self,) + args, kwargs)
        self.will_close = 0
        self.chunked = 0

    def begin(self):
        packet = AJPPacket()
        body = ""
        while 1:
            packet.receive(self.fp)
            type = packet.read_byte()

            if type == AJP_PACKET_SEND_HEADERS:
                if self.msg is not None:
                    raise AJP13Error, "multiple SEND_HEADERS received"

                self.status = packet.read_int()
                self.reason = packet.read_string()
                self.msg = AJP13Message(packet)        
                length = self.msg.getheader("content-length", None)
                if length is not None:
                    self.length = int(length)

            elif type == AJP_PACKET_SEND_BODY_CHUNK:
                body += packet.read_string()

            elif type == AJP_PACKET_END_RESPONSE:
                self.will_close != packet.read_byte()
                break

            else:
                raise AJP13Error, "unknown packet type %d" % type

        if self.length == httplib._UNKNOWN:
            self.length = len(body)

        self.fp = StringIO.StringIO(body)


class AJP13Connection(httplib.HTTPConnection):
    
    response_class = AJP13Response
    default_port = AJP_PORT

    def _send_output(self):
        req_line, headers = self._buffer[0], self._buffer[1:]
        del self._buffer[:]

        method, req_uri, protocol = req_line.split(" ")
        method = AJP_METHODS.index(method)
        req_uri = req_uri.split("?", 1)
        if len(req_uri) == 2:
            req_uri, query_string = req_uri
        else:
            req_uri, query_string = req_uri[0], ""

        # XXX should we pretend to be fronting for another server?
        if self.sock is None:
            if self.auto_open:
                self.connect()
            else:
                raise httplib.NotConnected()
        remote_addr = self.sock.getsockname()[0]
        remote_host = "" # XXX mod_jk2 does this

        is_ssl = 0

        # XXX not strictly necessary, but mod_jk2 always does this
        for header in headers:
            if header.split(": ")[0].lower() == "content-length":
                break
        else:
            headers.append("Content-length: 0")

        packet = AJPPacket()
        packet.extend((
            Byte(AJP_PACKET_FORWARD_REQUEST),
            Byte(method),
            protocol,
            req_uri,
            remote_addr,
            remote_host,
            self.host,
            self.port,
            Byte(is_ssl),
            len(headers)))

        for header in headers:
            name, value = header.split(": ")
            try:
                name = 0xA000 | AJP_HEADERS.index(name.lower())
            except ValueError:
                pass
            packet.extend((name, value))

        if query_string:
            packet.extend((Byte(AJP_ATTRIBUTE_QUERY_STRING), query_string))

        packet.append(Byte(0xFF))
        self.send(packet)

    def send(self, packet):
        if type(packet) == type(""):
            p = AJPPacket()
            p.append(packet)
            p.data = p.data[:-1]
            packet = p
        class FakeSock:
            """Wrapper to allow us to take advantage of the funky
            error recovery stuff in httplib.HTTPConnection.send()"""
            def __init__(self, conn):
                self.conn = conn
            def sendall(self, str):
                return apply(httplib.HTTPConnection.send, (self.conn, str))
        packet.transmit(FakeSock(self))
