import ajp13lib
import base64
import copy
import httplib
import log
import mimetools
import posixpath
import cStringIO as StringIO
import thread
import urlparse

__all__ = ["ClientSesson", "ClientResponse"]

# Make urlparse recognise the ajp13 URLs that Server.getSession() emits.
for list in [a for a in [getattr(urlparse, a) for a in dir(urlparse)]
             if type(a) == type([]) and "http" in a]:
    list.insert(list.index("http"), "ajp13")

# Only log messages from the main thread, which saves filling the
# logfile with crap during load tests.
main_thread = thread.get_ident()

           
##############################################################################
# A HTTP (or HTTP-like) session
##############################################################################

class ClientSession:
    """Class for managing HTTP and AJP 1.3 sessions with support for
    relative URLs, authentication and cookies"""

    def __init__(self, url):
        self.url = url
        self.cookies = []
        self.credentials = (None, None)

    def resolveURL(self, url):
        """Resolve a URL relative to the session, if necessary."""
        thisbits = urlparse.urlparse(url)
        if thisbits[:2] != ("", ""):
            return url
        basebits = urlparse.urlparse(self.url)
        if thisbits[2] == "":
            path = basebits[2]
        else:
            path = basebits[2]
            if not path.endswith("/"):
                path = posixpath.split(path)[0]
            path = posixpath.join(path, thisbits[2])
        return urlparse.urlunparse(basebits[:2] + (path,) + thisbits[3:])

    def setURL(self, url):
        """Set the base URL for subsequent requests."""
        self.url = self.resolveURL(url)
        self.urlbits = None

    def authenticate(self, user, password):
        """Authenticate subsequent requests with the specified credentials."""
        self.credentials = (user, password)

    def cookiesMatch(self, *expected):
        """Compare our cookies with a list of expected ones."""
        if len(expected) != len(self.cookies):
            return 0
        for expected, actual in zip(expected, self.cookies):
            if expected != (actual["name"], actual["value"]):
                return 0
        return 1

    # Methods for obtaining responses

    def GET(self, url, headers = None):
        return self._open("GET", url, headers, None, None)

    def POST(self, url, data, type = "application/x-www-form-urlencoded",
             headers = None):
        return self._open("POST", url, headers, data, type)

    def PUT(self, url, data, type, headers = None):
        return self._open("PUT", url, headers, data, type)

    # Make a request and obtain a response

    def _open(self, method, url, headers, data, type):
        url = self.resolveURL(url)
        if thread.get_ident() == main_thread:
            log.debug('%s "%s"' % (method, url))

        scheme, netloc, path, params, query, fragment = urlparse.urlparse(url)
        request_uri = urlparse.urlunparse(("", "", path, params, query, ""))

        # Build a list of headers to send
        reqheaders = HeaderSet()
        if headers is not None:
            for name, value in headers:
                reqheaders[name] = value
        
        host = netloc.split(":", 1)[0]
        cookies = {}
        for cookie in self.cookies:
            if (host.endswith(cookie["domain"]) and 
                path.startswith(cookie["path"])):
                cookies[cookie["name"]] = cookie["value"]
        cookies = "; ".join(["%s=%s" % (k, cookies[k]) for k in cookies])
        if cookies:
            if thread.get_ident() == main_thread:
                log.debug('with cookies "%s"' % cookies)
            reqheaders["Cookie"] = cookies

        if self.credentials[0] is not None:
            reqheaders["Authorization"] = "Basic %s" % base64.encodestring(
                "%s:%s" % self.credentials).strip()

        if data is not None:
            reqheaders["Content-length"] = str(len(data))
            reqheaders["Content-type"] = type

        # Perform the request
        conn = pool.getConnection(scheme, netloc)
        conn.putrequest(method, request_uri)
        for name, value in reqheaders.items():
            conn.putheader(name, value)
        # XXX evil, but it allows us to get exactly the headers that were sent.
        reqheaders = mimetools.Message(StringIO.StringIO("\n".join(
            conn._buffer[1:])))
        reqheaders.unixfrom = conn._buffer[0]
        # XXX end of evil
        conn.endheaders()
        if data is not None:
            conn.send(data)
        response = conn.getresponse()

        # Update the session with cookies and stuff
        self.url = url
        host, path = urlparse.urlparse(url)[1:3]
        host = host.split(":", 1)[0]
        for cookie in response.msg.getallmatchingheaders("Set-Cookie"):
            parts = map(lambda s: s.strip(),
                        cookie.split(":", 1)[1].split(";"))
            cookie = parts.pop(0)
            attrs = {"domain": host, "path": path}
            for part in parts:
                key, value = part.split("=", 1)
                attrs[key.lower()] = value
            name, value = cookie.split("=", 1)
            self.cookies.append({"name":   name,
                                 "value":  value,
                                 "domain": attrs["domain"],
                                 "path":   attrs["path"]})

        # Build and return the response
        return ClientResponse(reqheaders,
                              response.status, response.reason, response.msg,
                              response.read())


class ClientResponse:

    def __init__(self, reqheaders, status, reason, headers, body):
        # The request that triggered this response
        self.reqheaders = reqheaders

        # The response itself
        self.status = status
        self.reason = reason
        self.headers = headers
        self.type = headers.type
        self.body = body
        self.lines = body.replace("\r", "").split("\n")
        if self.lines[-1] == "":
            self.lines.pop()

    def readline(self):
        if self.lines:
            return self.lines.pop(0)
        else:
            return None


class HeaderSet:
    """Class to manage a set of headers.  Think of it as a dictionary
    with case-insensitive keys that retains the order that things were
    added and rejects duplicate assignments."""
    
    def __init__(self):
        self.data = {}
        self.order = []

    def __getitem__(self, key):
        return self.data[key.lower()]

    def __setitem__(self, key, value):
        if self.data.has_key(key.lower()):
            raise KeyError, "%s exists" % key
        self.data[key.lower()] = value
        self.order.append(key)

    def items(self):
        return [(key, self.data[key.lower()]) for key in self.order]


##############################################################################
# Connection pooling
##############################################################################

class ConnectionPool:
    """Thread-safe pool of connections."""

    def __init__(self):
        self.mutex = thread.allocate_lock()
        self.pool  = {}

    def getConnection(self, scheme, netloc):
        key = scheme + ":" + netloc
        self.mutex.acquire()
        if self.pool.has_key(key):
            list = self.pool[key]
        else:
            list = self.pool[key] = []
        if list:
            conn = list.pop()
        else:
            conn = {"http": httplib.HTTPConnection,
                    "ajp13": ajp13lib.AJP13Connection}[scheme](netloc)
        self.mutex.release()
        return ConnectionWrapper(conn, pool, key)

    def returnConnection(self, conn, key):
        self.mutex.acquire()
        self.pool[key].append(conn)
        self.mutex.release()


class ConnectionWrapper:
    """Wrapper for connections to ensure that they are returned to the
    pool after use."""

    def __init__(self, conn, pool, key):
        self.__conn, self.__pool, self.__key = conn, pool, key

    def __del__(self):
        self.__pool.returnConnection(self.__conn, self.__key)
        
    def __getattr__(self, attr):
        return getattr(self.__conn, attr)


pool = ConnectionPool()


##############################################################################
# Test stuff
##############################################################################

if __name__ == "__main__":
    print ClientSession(
        "ajp13://localhost:8009/examples/").GET("jsp/num/numguess.jsp").body
    print ClientSession(
        "ajp13://localhost:8009/examples/").GET("jsp/num/numguess.jsp").body
