import clientsession
import config
import log
import os
import signal
import smtplogger
import sys
import time
import urlparse
import util
import weakref

__all__ = ["catalina_home", "template_home", "executable", "spawn",
           "getServer"]

# Parameters that the caller may set to override the defaults
catalina_home = "/tmp/tomcat"
template_home = "/usr/share/tomcat"
executable = "/usr/sbin/tomcat"
spawn = 1


##############################################################################
# A running Tomcat server
##############################################################################

class ServerError(Exception):
    """The server failed to start."""

# Path of script to create catalina home
if __name__ == "__main__":
    mkhome_script = sys.argv[0]
else:
    mkhome_script = __file__
mkhome_script = os.path.join(
    os.path.split(os.path.realpath(mkhome_script))[0], "mkcatalinahome.sh")

class Server:
    """Class through which everything interacts with the server."""

    manageruser = "manageruser"
    managerpass = "managerpass"

    adminuser = "adminuser"
    adminpass = "adminpass"

    def __init__(self, connector = "jk2"):
        # Make sure these are available when the destructor runs
        self.log, self.os, self.signal = log, os, signal

        self.connector = connector

        self.dir = catalina_home
        self.logfile = os.path.join(self.dir, "logs", "tomcat.out")
        self.pidfile = os.path.join(self.dir, "tomcat.pid")
        self.pid = None

        if spawn:
            self.makeHome()
            self.start()
        else:
            try:
                self.pid = int(open(self.pidfile, "r").readline().strip())
            except:
                pass

        s = self.getSession("manager")
        s.authenticate("manageruser", "managerpass")
        self.manager = Manager(s)

    def __del__(self):
        if spawn and self.pid is not None:
            self.stop()

    def makeHome(self):
        log.info("creating catalina home in " + self.dir)
        util.system("/bin/bash", mkhome_script, self.dir, template_home)
        config.write(open(os.path.join(self.dir, "conf", "server.xml"), "w"),
                     {"connector": self.connector,
                      "realm": "file",
                      "examples": "true",
                      "test": "true"})
        open(os.path.join(self.dir, "conf", "tomcat-users.xml"), "w").write(
            "<tomcat-users>\n" +
            '  <user name="%s" password="%s" roles="manager" />\n' % (
                    self.manageruser, self.managerpass) +
            '  <user name="%s" password="%s" roles="admin" />\n' % (
                    self.adminuser, self.adminpass) +
            "</tomcat-users>\n")

    def start(self):
        log.info("starting server")
        startTime = time.time()
        util.system(executable,
                    "-home", self.dir,
                    "-logfile", self.logfile,
                    "-pidfile", self.pidfile)
        self.pid = int(open(self.pidfile, "r").readline().strip())

        log.debug("waiting for server to warm up")
        ready, broken = {
            "http": ("Coyote HTTP/1.1 on port 8080",
                     "Address already in use:8080"),
            "jk2": ("ajp13 listening on localhost.localdomain/127.0.0.1:8009",
                    "Address already in use")}[self.connector]
        for i in range(600):
            for line in open(self.logfile, "r").xreadlines():
                if line.strip().endswith(ready):
                    break
                elif line.strip().endswith(broken):
                    raise ServerError, "address already in use"
            else:
                time.sleep(0.1)
                continue
            break
        else:
            raise ServerError, "gave up waiting for server to warm up"

        stopTime = time.time()
        timeTaken = stopTime - startTime

        log.info("server started and warmed up in %.3f seconds" % timeTaken)

    def stop(self):
        self.log.info("stopping server")
        self.os.kill(self.pid, self.signal.SIGTERM)
        self.pid = None

    def rotateLogs(self):
        log.info("rotating logfiles")
        os.rename(self.logfile, self.logfile + ".1")
        os.kill(self.pid, signal.SIGUSR1)

    def getSession(self, context):
        """Return a new session through which requests may be made."""
        return clientsession.ClientSession({
            "http": "http://localhost:8080/%s/",
            "jk2": "ajp13://localhost:8009/%s/"}[self.connector] % context)

    def logMailSession(self, callable):
        """Log the SMTP transaction that occurs when callable is invoked"""
        return smtplogger.logSMTPTransaction("localhost", 8025, callable)

# Function through which tests obtain a running server
server = None

def getServer(*args, **kwargs):
    """Return a running Tomcat server, creating one if necessary."""
    global server
    if server is None:
        server = apply(Server, args, kwargs)
    elif args or kwargs:
        raise ServerError, "server is already running"
    return weakref.proxy(server)


##############################################################################
# Manage the webapps on a server
##############################################################################

class ManagerError(Exception):
    """An interaction with the Manager application failed."""

class Manager:
    """Class through which tests access the Manager application."""

    def __init__(self, session):
        self.session = session

    def install(self, app):
        scheme, netloc, path, params, query, fragment = urlparse.urlparse(app)
        if path.endswith(".xml"):
            if not scheme:
                scheme = "file"
            if scheme == "file" and not os.path.isabs(path):
                path = os.path.abspath(path)
        app = urlparse.urlunparse((
            scheme, netloc, path, params, query, fragment))
        return apply(self.process, ("install",),
              {path.endswith(".xml") and "config" or "war": app})

    def __getattr__(self, attr):
        class Helper:
            def __init__(self, action, manager):
                self.action = action
                self.manager = manager
            def __call__(self, path, **args):
                return apply(self.manager.process, (self.action, path), args)
        return Helper(attr, self)

    def process(self, action, path = None, **args):
        if path:
            if args.has_key("path"):
                raise ManagerError, "path specified twice"
            args["path"] = path
        response = self.session.GET(action + "?" + "&".join(
            ["%s=%s" % (arg, args[arg]) for arg in args]))
        if response.status != 200:
            raise ManagerError, \
                  "request failed with status %d" % response.status

        msg = response.lines[0].rstrip()
        code, response = msg.split(" - ", 1)
        if code != "OK":
            raise ManagerError, msg
        log.debug(response[0].lower() + response[1:])

    def serverinfo(self):
        response = self.session.GET("serverinfo")
        if response.status != 200:
            raise ManagerError, \
                  "request failed with status %d" % response.status
        info = {}
        for line in response.lines:
            key, value = line.split(": ", 1)
            info[key.lower().replace(" ", "-")] = value
        return info
