import copy
import glob
import log
import os
import server
import sys
import time
import unittest

# Parameter that the caller may set to override the default
if __name__ == "__main__":
    testdir = sys.argv[0]
else:
    testdir = __file__
testdir = os.path.join(os.path.split(os.path.split(
    os.path.realpath(testdir))[0])[0], "tests")


##############################################################################
# Testcases
##############################################################################

class TestSkippedError(Exception):
    """Thrown to indicate skipped tests."""
    def __init__(self, exc_info):
        self.exc_info = exc_info

class TestCase(unittest.TestCase):
    """Base class for all tests that require a running Tomcat server."""

    webapp = None
    deploy = 0
    order = 50

    skippedException = TestSkippedError

    def setUp(self):
        if self.webapp is not None:
            self.server = server.getServer()
            if self.deploy:
                self.server.manager.install(
                    os.path.join(self.getDataDir(), self.webapp + ".xml"))

    def tearDown(self):
        if self.deploy:
            self.server.manager.remove("/" + self.webapp)

    def skip(self, exc_info = None):
        raise self.skippedException(exc_info)

    # Functions for interacting with the server

    def getSession(self, webapp = None):
        """Return a new HTTP session through which requests may be made."""
        return self.server.getSession(webapp is None and self.webapp or webapp)

    def logMailSession(self, callable):
        """Log the SMTP transaction that occurs when callable is invoked"""
        return self.server.logMailSession(callable)

    def getAppsDir(self):
        """The directory in which the server looks for webapps."""
        return os.path.join(self.server.dir, "webapps")

    def getDataDir(self):
        """A directory in which data can be stored for tests."""
        file = sys.modules[self.__module__].__file__
        if file.endswith(".py"):
            return file[:-3]
        elif file[:-1].endswith(".py"):
            return file[:-4]
        else:
            raise ValueError, "bad module filename %s" % file


class TestSuite(unittest.TestSuite):
    """A composite test consisting of a number of TestCases."""

    # Some serious hoop-jumping in order to propagate the sort order
    # from where it is defined to where it is required.
    __order1 = None
    __order2 = None

    def addTest(self, test):
        if isinstance(test, TestCase):
            if self.__order1 is None:
                self.__order1 = test.order
            elif self.__order1 != test.order:
                raise ValueError, "test order mismatch"
        elif isinstance(test, TestSuite):
            if self.__order2 is None:
                self.__order2 = test.__order1
            elif test.__order1 is not None:
                if self.__order2 != test.__order1:
                    raise ValueError, "suite order mismatch"
        return apply(unittest.TestSuite.addTest, (self, test))

    def __call__(self, result):
        if self._tests and isinstance(self._tests[0], TestSuite):
            if self._tests[0].__order2 is not None:
                self._tests.sort(lambda a, b: cmp(a.__order2, b.__order2))
        return apply(unittest.TestSuite.__call__, (self, result))


##############################################################################
# Text-based UI with logging
##############################################################################

class TestResult(unittest._TextTestResult):
    """A test result class that prints formatted results if to a tty
    or logs them otherwise."""
    separator1 = '=' * 80
    separator2 = '-' * 80

    def __init__(self, *args, **kwargs):
        apply(unittest._TextTestResult.__init__, (self,) + args, kwargs)
        self.skipped = []

    def startTest(self, test):
        unittest._TextTestResult.startTest(self, test)
        if self.stream.isatty():
            print >>self.stream, " %-70s" % test,
            self.stream.flush()
        else:
            log.info("running %s" % test)

    def addSuccess(self, test):
        unittest._TextTestResult.addSuccess(self, test)
        if self.stream.isatty():
            print >>self.stream, " [\x1B[32mPASS\x1B[0m]"

    def addFailure(self, test, err):
        unittest._TextTestResult.addFailure(self, test, err)
        if self.stream.isatty():
            print >>self.stream, " [\x1B[1;31mFAIL\x1B[0m]"
        else:
            log.warn("%s failed" % test)

    def addError(self, test, err):
        if err[0] is test.skippedException:
            self.addSkipped(test, err[1].exc_info)
        else:
            unittest._TextTestResult.addError(self, test, err)
            if self.stream.isatty():
                print >>self.stream, "[\x1B[1;33mERROR\x1B[0m]"
            else:
                log.warn("%s is broken" % test)

    def addSkipped(self, test, err):
        self.skipped.append((test, self._exc_info_to_string(err)))
        if self.stream.isatty():
            print >>self.stream, " [\x1B[1;36mSKIP\x1B[0m]"
        else:
            log.warn("%s was skipped" % test)
        

class TestRunner(unittest.TextTestRunner):
    """A test runner class that prints formatted results if to a tty
    or logs them otherwise."""

    def __init__(self):
        unittest.TextTestRunner.__init__(self, log.stream)
    
    def run(self, test):
        result = TestResult(self.stream, 1, 0)
        if self.stream.isatty():
            self.stream.write(result.separator2 + "\n")

        startTime = time.time()
        test(result)
        stopTime = time.time()
        timeTaken = float(stopTime - startTime)

        if not result.wasSuccessful():
            self.stream.write("\n")
        result.printErrors()
        if not log.loglevel:
            result.printErrorList("SKIPPED", result.skipped)
        
        if self.stream.isatty():
            self.stream.write(result.separator2 + "\n")
        log.info("ran %d test%s in %.3f seconds" % (
            result.testsRun, result.testsRun == 1 and "" or "s", timeTaken))

        if result.wasSuccessful():
            log.info("all tests passed")
        else:
            failed, errored = map(len, (result.failures, result.errors))
            msg = "FAILED ("
            if failed:
                msg += "failures=%d" % failed
            if errored:
                if failed:
                    msg += ", "
                msg += "errors = %d" % errored
            msg += ")"
            log.warn(msg)

        if result.skipped:
            log.warn("%d test%s skipped" % (
                len(result.skipped),
                len(result.skipped) == 1 and " was" or "s were"))

        return result


##############################################################################
# Load tests from a specific directory
##############################################################################

class LoaderError(Exception):
    """The loader was unable to load any tests"""

class TestLoader(unittest.TestLoader):
    """Load tests according to various criteria."""
    suiteClass = TestSuite

    def loadDefaultTests(self):
        """Return a suite of all test cases contained in testdir."""
        files = glob.glob(os.path.join(testdir, "*.py"))
        if not files:
            raise LoaderError, "no testcases found"
        return self.loadTestsFromPaths(files)

    def loadTestsFromPaths(self, files):
        """Return a suite of tests loaded from the specified list of files."""
        files = map(os.path.abspath, files)
        paths = copy.copy(sys.path)
        paths.sort(lambda a, b: cmp(len(b), len(a)))
        for path in paths:
            path += os.sep
            if path and os.path.commonprefix([path] + files) == path:
                break
        else:
            raise LoaderError, "testcases must reside in sys.path"

        return self.loadTestsFromNames(
            [f[len(path):-3].replace(os.sep, ".")
             for f in files
             if not f.endswith(os.sep + "__init__.py")])

    def getTestCaseNames(self, testCaseClass):
        if hasattr(testCaseClass, "getTestNames"):
            # Create a dummy instance so stuff like getDataDir works.
            # We need to pass an existing method name to the
            # constructor, so we use getTestNames since we just
            # checked that it exists.
            testFnNames = testCaseClass("getTestNames").getTestNames()
            if self.sortTestMethodsUsing:
                testFnNames.sort(self.sortTestMethodsUsing)
            return testFnNames
        else:
            return unittest.TestLoader.getTestCaseNames(self, testCaseClass)
        

##############################################################################
# Facilities for running tests
##############################################################################

def runTests(tests = None):
    loader = TestLoader()
    if tests:
        suite = loader.loadTestsFromPaths(tests)
    else:
        suite = loader.loadDefaultTests()
    return TestRunner().run(suite).wasSuccessful()

if __name__ == "__main__":
    sys.path.insert(0, os.path.split(testdir)[0])
    runTests()
