# Tomcat Example Application test
# Coverage: Pretty much complete.
# Exercises: JSPs, Tag libraries, Sessions, Cookies, JNDI, JavaMail.

from harness.testsuite import TestCase

import re
import rfc822
import cStringIO as StringIO
import socket
import sys
import time
import urlparse

class ExamplesTestCase(TestCase):
    webapp = "examples"

    # The Hello World example servlet is about the simplest a servlet
    # can be, so we check just about everything it does.
    def testHelloWorld(self):
        s = self.getSession()
        r = s.GET("servlet/HelloWorldExample")
        self.assertEqual(r.status, 200)
        self.failUnless(s.cookiesMatch())
        self.assertEqual(r.type, "text/html")
        self.failUnless(re.search(
            r"^<h1>Hello World!</h1>\s*$", r.body, re.M))

    # The Request Info example servlet dumps various bits and pieces
    # from the request, so we check them to ensure they are getting
    # through correctly.
    def testRequestInfo(self):
        s = self.getSession()
        baseurl = s.resolveURL("servlet/RequestInfoExample")
        for method in ("GET", "POST"):
            for pathinfo in (None, "/foo"):
                url = baseurl
                if pathinfo:
                    url += pathinfo
                else:
                    pathinfo = "null"
                r = getattr(s, method)(url, {"GET": None, "POST": ""}[method])
                self.assertEqual(r.status, 200)

                request_uri = urlparse.urlunparse(
                    ("", "") + urlparse.urlparse(s.url)[2:4] + ("", ""))

                expected = {
                    "Method": method,
                    "Path Info": pathinfo,
                    "Request URI": request_uri,
                    "Remote Address": "127.0.0.1"}

                actual = {}
                state = 0
                for line in r.lines:
                    line = line.rstrip()
                    if state == 0:
                        if line.startswith("<table"):
                            state = 1
                    elif state == 1:
                        self.failUnless(line.endswith(":"))
                        key = line[:-1]
                        state = 2
                    elif state == 2:
                        self.assertEqual(line, "</td><td>")
                        state = 3
                    elif state == 3:
                        actual[key] = line
                        del key
                        state = 4
                    elif state == 4:
                        if line == "</table>":
                            state = 5
                        else:
                            self.assertEqual(line, "</td></tr><tr><td>")
                            state = 1
                self.assertEqual(state, 5)

                for key in expected:
                    self.failUnless(actual.has_key(key))
                    self.assertEqual(expected[key], actual[key])

    # The Request Headers example servlet dumps the headers we are
    # sending.
    def testRequestHeaders(self):
        s = self.getSession()
        s.setURL("servlet/RequestHeaderExample")
        for method in ("GET", "POST"):
            for headers in ([], [["Funky", "Monkey"]]):
                if method == "GET":
                    r = s.GET("", headers)
                else:
                    r = s.POST("", "monkey")
                self.assertEqual(r.status, 200)

                expected = r.reqheaders.dict
                actual = {}
                state = 0
                for line in r.lines:
                    line = line.rstrip()
                    if state == 0:
                        if line.startswith("<table"):
                            state = 1
                    elif state == 1:
                        if line == "</table>":
                            state = 6
                        else:
                            self.assertEqual(line,
                                             '<tr><td bgcolor="#CCCCCC">')
                            state = 2
                    elif state == 2:
                        key = line.lower()
                        state = 3
                    elif state == 3:
                        self.assertEqual(line, "</td><td>")
                        state = 4
                    elif state == 4:
                        actual[key] = line
                        del key
                        state = 5
                    elif state == 5:
                        self.assertEqual(line, "</td></tr>")
                        state = 1
                self.assertEqual(state, 6)

                for key in expected:
                    self.failUnless(actual.has_key(key))
                    self.assertEqual(expected[key], actual[key])

    # The Request Parameters example servlet dumps the request parameters.
    def testRequestParameters(self):
        s = self.getSession()
        r = s.GET("servlet/RequestParamExample")
        self.assertEqual(r.status, 200)
        self.failUnless(re.search(
            r"^No Parameters, Please enter some\s*$", r.body, re.M))

        r = s.GET("?firstname=funky&lastname=monkey")
        self.assertEqual(r.status, 200)
        self.failUnless(re.search(
            r"^Parameters in this request:<br>\s+" +
            r"^First Name:\s+=\s+funky<br>\s+Last Name:\s+=\s+monkey",
            r.body, re.M))

        r = s.POST("", "firstname=colin&lastname=cowpat")
        self.assertEqual(r.status, 200)
        self.failUnless(re.search(
            r"^Parameters in this request:<br>\s+" +
            r"^First Name:\s+=\s+colin<br>\s+Last Name:\s+=\s+cowpat",
            r.body, re.M))

    # The Cookies example servlet sets and dumps cookies.
    def testCookies(self):
        s = self.getSession()
        r = s.GET("servlet/CookieExample")
        self.assertEqual(r.status, 200)
        self.failUnless(s.cookiesMatch())
        self.failUnless(re.search(
            r"^Your browser isn't sending any cookies\s*$", r.body, re.M))

        r = s.GET("?cookiename=moose&cookievalue=hoose")
        self.assertEqual(r.status, 200)
        self.failUnless(s.cookiesMatch(("moose", "hoose")))
        self.failUnless(re.search(
            r"^Your browser isn't sending any cookies\s+" +
            r"^<P>\s+" +
            r"^You just sent the following cookie to your browser:<br>\s+" +
            r"^Name:\s+moose<br>Value:\s+hoose<P>", r.body, re.M))

        r = s.POST("", "cookiename=TheCat&cookievalue=InTheHat")
        self.assertEqual(r.status, 200)
        self.failUnless(s.cookiesMatch(
            ("moose", "hoose"), ("TheCat", "InTheHat")))
        self.failUnless(re.search(
            r"^Your browser is sending the following cookies:<br>\s+" +
            r"^Cookie Name:\s+moose<br>\s+Cookie Value:\s+hoose<br><br>\s+" +
            r"^<P>\s+" +
            r"^You just sent the following cookie to your browser:<br>\s+" +
            r"^Name:\s+TheCat<br>Value:\s+InTheHat<P>", r.body, re.M))

        r = s.GET("?cookiename=TheCat&cookievalue=Stinks")
        self.assertEqual(r.status, 200)
        self.failUnless(s.cookiesMatch(
            ("moose", "hoose"), ("TheCat", "InTheHat"), ("TheCat", "Stinks")))
        self.failUnless(re.search(
            r"^Your browser is sending the following cookies:<br>\s+" +
            r"^Cookie Name:\s+TheCat<br>\s+" +
            r"Cookie Value:\s+InTheHat<br><br>\s+" +
            r"^Cookie Name:\s+moose<br>\s+Cookie Value:\s+hoose<br><br>\s+" +
            r"^<P>\s+" +
            r"^You just sent the following cookie to your browser:<br>\s+" +
            r"^Name:\s+TheCat<br>Value:\s+Stinks<P>", r.body, re.M))

    # The Naming example servlet demonstrates JNDI lookups.
    def testNaming(self):
        r = self.getSession().GET("servlet/JndiServlet")
        self.assertEqual(r.status, 200)
        self.assertEqual(r.type, "text/plain")

        self.failUnless(r.readline().startswith(
            "Simple lookup test :"))
        self.failUnless(r.readline().startswith(
            "Max exemptions value : 15"))  # configured in web.xml
        self.failUnless(r.readline().startswith(
            "Resource link test :"))
        self.failUnless(r.readline().startswith(
            "Link value : 30")) # configured in server.xml

        self.failUnless(r.readline().startswith(
            "list() on /comp/env Context :"))
        names = []
        while 1:
            line = r.readline()
            if line.startswith("Binding : "):
                names.append(line[10:].rstrip())
            else:
                break
        self.failUnless(line.startswith(
            "listBindings() on /comp/env Context :"))
        values = {}
        for name in names:
            expected = "Binding : " + name + ":"
            line = r.readline()
            self.failUnless(line.startswith(expected))
            values[name] = line[len(expected):].rstrip()

        expected = {"minExemptions:java.lang.Integer": 1,
                    "maxExemptions:java.lang.Integer": 15,
                    "name3:java.lang.Integer": 1,
                    "linkToGlobalResource:java.lang.Integer": 30}
        for key in expected:
            self.failUnless(values.has_key(key))
            self.assertEqual(expected[key], int(values[key]))
        for key in values:
            if not expected.has_key(key):
                self.failUnless(values[key].startswith(
                    "org.apache.naming.NamingContext@"))

        self.assertEqual(len(r.lines), 0)

    # The Sessions example servlet demonstrates sessions.  We check
    # that the session is created, and that the server sets the cookie
    # and urlencodes the session id where appropriate.
    def testSessions(self):
        s = self.getSession()
        r = s.GET("servlet/SessionExample")
        self.assertEqual(r.status, 200)
        self.assertEqual(len(s.cookies), 1)
        cookie = s.cookies[0]
        self.assertEqual(cookie["name"], "JSESSIONID")
        session = cookie["value"]
        self.failUnless(re.search(
            r"^Session ID: %s" % session +
            r".*" +
            r"^The following data is in your session:<br>\s+" +
            r"^<P>\s+" +
            r'^<form action="SessionExample;jsessionid=%s"' % session,
            r.body, re.M | re.S))

        r = s.GET("")
        self.assertEqual(r.status, 200)
        self.failUnless(s.cookiesMatch(("JSESSIONID", session)))
        self.failUnless(re.search(
            r"^Session ID: %s" % session +
            r".*" +
            r"^The following data is in your session:<br>\s+" +
            r"^<P>\s+" +
            r'^<form action="SessionExample', r.body, re.M | re.S))
        
        r = s.GET("?dataname=monkey&datavalue=nuts")
        self.assertEqual(r.status, 200)
        self.failUnless(s.cookiesMatch(("JSESSIONID", session)))
        self.failUnless(re.search(
            r"^Session ID: %s" % session +
            r".*" +
            r"^The following data is in your session:<br>\s+" +
            r"^monkey = nuts<br>\s+" +
            r"^<P>\s+" +
            r'^<form action="SessionExample', r.body, re.M | re.S))
        
        r = s.POST("", "dataname=I like&datavalue=Bananas in Pyjamas")
        self.assertEqual(r.status, 200)
        self.failUnless(s.cookiesMatch(("JSESSIONID", session)))
        self.failUnless(re.search(
            r"^Session ID: %s" % session +
            r".*" +
            r"^The following data is in your session:<br>\s+" +
            r"^I like = Bananas in Pyjamas<br>\s+" +
            r"^monkey = nuts<br>\s+" +
            r"^<P>\s+" +
            r'^<form action="SessionExample', r.body, re.M | re.S))

        s.cookies = []
        r = s.GET(";jsessionid=%s" % session)
        self.assertEqual(r.status, 200)
        self.failUnless(s.cookiesMatch())
        self.failUnless(re.search(
            r"^Session ID: %s" % session +
            r".*" +
            r"^The following data is in your session:<br>\s+" +
            r"^I like = Bananas in Pyjamas<br>\s+" +
            r"^monkey = nuts<br>\s+" +
            r"^<P>\s+" +
            r'^<form action="SessionExample;jsessionid=%s"' % session,
            r.body, re.M | re.S))

    # The Numberguess example JSP uses a session-scoped bean, and
    # hence exercises sessions and cookies and the like.  We check
    # that the session is created correctly, and that it is never
    # overwritten.
    def testNumberGuess(self):
        s = self.getSession()
        r = s.GET("jsp/num/numguess.jsp")
        self.assertEqual(r.status, 200)
        self.assertEqual(len(s.cookies), 1)
        cookie = s.cookies[0]
        self.assertEqual(cookie["name"], "JSESSIONID")
        session = cookie["value"]
        m = re.search(
            r"Welcome to the Number Guess game\.<p>" +
            r"\s+" +
            r"I'm thinking of a number between (\d+) and (\d+)\.<p>",
            r.body, re.S)
        self.failUnless(m)
        min, max = map(int, m.group(1, 2))
        range = "%d and %d" % (min, max)

        min, max, guesses = min - 1, max + 1, 0
        while 1:
            step = (max - min) / 2
            guess = min + step
            guesses += 1

            r = s.GET("?guess=%d" % guess)
            self.assertEqual(r.status, 200)
            self.failUnless(s.cookiesMatch(("JSESSIONID", session)))
            m = re.search(
                r"(Good guess, but nope\.  Try <b>(higher|lower)</b>\." +
                r"\s+" +
                r"You have made %d guesses\.<p>" % guesses +
                r"\s+" +
                r"I'm thinking of a number between %s\.<p>)|" % range +
                r"(Congratulations!  You got it\." +
                r"\s+" +
                r"And after just %d tries\.<p>)" % guesses,
                r.body, re.S)
            self.failUnless(m)

            result = m.group(2)
            if result == "higher":
                min = guess
            elif result == "lower":
                max = guess
            else:
                break

    # The Date example JSP uses a page-scoped bean, and explicitly
    # doesn't use a session.  We check that a session is not created,
    # and that the date information returned is consistent.
    def testDate(self):
        s = self.getSession()
        r = s.GET("jsp/dates/date.jsp")
        self.assertEqual(r.status, 200)
        self.failUnless(s.cookiesMatch())

        data = {}
        ex = re.compile(r"^<li>\s*(\w.+): is\s+(\S+)$")
        for line in r.lines:
            m = ex.search(line)
            if m:
                key, value = m.group(1, 2)
                data[key] = value
        self.assertEqual(len(data), 11)

        tuple = time.strptime(
            data["Date"] + " " + data["Time"], "%m/%d/%Y %H:%M:%S")

        # Mess with week of year.  The first day of the year is always
        # in week one in Java, but in Python week one is the first
        # _complete_ week.
        if int(time.strftime("%W",
                             time.strptime("%04d-01-01" % tuple[0],
                                           "%Y-%m-%d"))) == 0:
            data["Week Of Year"] = str(int(data["Week Of Year"]) - 1)

        tests = {"Year": ("%Y", int), "Month": ("%B", str), "Day": ("%A", str),
                 "Day Of Year": ("%j", int), "Week Of Year": ("%W", int)}
        for key in tests:
            format, type = tests[key]
            apply(self.assertEqual, map(type, (
                data[key], time.strftime(format, tuple))))

    # The Snoop example JSP dumps various bits and pieces from the
    # request, so we check them to ensure they are getting through
    # correctly.  They ought to be, since we checked them all for the
    # servlets earlier.
    def testSnoop(self):
        s = self.getSession()
        s.setURL("jsp/snp/snoop.jsp")
        for method in ("GET", "POST"):
            for url in ("", "?bar=baz"):
                r = getattr(s, method)(url, {
                    "GET": None, "POST": "monkey"}[method])
                self.assertEqual(r.status, 200)

                scheme, netloc, path, params, query, fragment = \
                    urlparse.urlparse(s.url)
                netloc = netloc.split(":", 1)
                if len(netloc) == 1:
                    host, port = netloc[0], "80"
                else:
                    host, port = netloc
                request_uri = urlparse.urlunparse(
                    ("", "", path, params, "", ""))

                expected = {
                    "JSP Request Method": method,
                    "Request URI": request_uri,
                    "Servlet path": "/jsp/snp/snoop.jsp",
                    "Query string": query or "null",
                    "Content length": r.reqheaders.getheader(
                        "Content-length", {
                            "jk2": "0", "http": "-1"}[self.server.connector]),
                    "Content type": r.reqheaders.getheader(
                        "Content-type", "null"),
                    "Server name": host,
                    "Server port": port,
                    "Remote address": "127.0.0.1"}

                actual = {}
                for line in r.lines:
                    bits = line.rstrip().split(": ", 1)
                    if len(bits) == 2:
                        actual[bits[0]] = bits[1]

                for key in expected:
                    self.failUnless(actual.has_key(key))
                    self.assertEqual(expected[key], actual[key])

    # The ErrorPage example JSP exercises the errorPage directive.
    def testErrorPage(self):
        s = self.getSession()
        s.setURL("jsp/error/err.jsp")
        for car in ("audi", "bmw328i", "infiniti", "integra", "z3"):
            r = s.GET("?name=" + car)
            self.assertEqual(r.status, 200)
            self.failUnless(re.search(
                r'^<body bgcolor="%s">' % {
                    "integra": "lightblue"}.get(car, "red"),
                r.body, re.M))

    # The Carts, Checkbox, Color and Calendar JSPs do nothing new.

    # The Include example JSP exercises the include and jsp:include
    # directives.
    def testInclude(self):
        r = self.getSession().GET("jsp/include/include.jsp")
        self.assertEqual(r.status, 200)
        count = 0
        time_re = re.compile("^(\d+)$")
        for line in r.lines:
            m = time_re.search(line.strip())
            if m:
                count += 1
                expected = time.mktime(time.localtime())
                actual = int(m.group(1)[:-3])
                self.failUnless(abs(expected - actual) < 5)
        self.assertEqual(count, 2)

    # The Forward JSP exercises the jsp:forward directive, which is
    # also exercised by the JSP-Servlet-JSP JSP example.

    # The Plugin JSP does nothing particularly interesting.

    # The JSP-Servlet-JSP example JSP exercises the jsp:forward
    # directive (and it's servlet equivalent).
    def testJSPServletJSP(self):
        r = self.getSession().GET("jsp/jsptoserv/jsptoservlet.jsp")
        self.assertEqual(r.status, 200)
        self.failUnless(re.search(
            r"^I have been invoked by\s+^servletToJsp\s+^Servlet.\s+",
            r.body, re.M))

    # The Simple Custom Tag example JSP demonstrates a pair of simple
    # custom tags.
    def testSimpleCustomTag(self):
        r = self.getSession().GET("jsp/simpletag/foo.jsp")
        self.assertEqual(r.status, 200)
        self.failUnless(re.search(
            r"^Radio stations that rock:\s+" +
            r"^<ul>\s+" +
            r"^<li>98\.5</li>\s+" +
            r"^<li>92\.3</li>\s+" +
            r"^<li>107\.7</li>\s+" +
            r"^</ul>\s+" +
            r"^Did you see me on the browser window as well?",
            r.body, re.M))

        log = [line for line in open(self.server.logfile, "r").readlines()]
        self.assertEqual("Did you see me on the stderr window?",
                         log[-5].strip())
        self.assertEqual("Did you see me on the browser window as well?",
                         log[-2].strip())

    # The Send Mail example JSP exercises JavaMail.
    def testSendMail(self):
        s = self.getSession()
        try:
            r, log = self.logMailSession(lambda: s.POST(
                "SendMailServlet",
                "mailfrom=Mr Magoo <mr.magoo@example.com>&" +
                "mailto=Mrs Magoo <mrs.magoo@example.com>&" +
                "mailsubject=Hello Mrs Magoo!&" +
                "mailcontent=What's for tea Mrs Magoo?\n" +
                "I hope it's cherry pie cos I like that a lot\n" +
                "Later!!!"))
        except socket.error:
            if str(sys.exc_value) != "(98, 'Address already in use')":
                raise
            self.skip(sys.exc_info())

        self.assertEqual(r.status, 200)
        self.failUnless(re.search(
            "Message successfully sent!", r.body))

        m = re.search(
            r"^220 localhost Service ready\n" +
            r"EHLO (\S+)\n" +
            r"502 Command not implemented\n" +
            r"HELO \1\n" +
            r"250 Requested mail action okay, completed\n" +
            r"MAIL FROM:<mr.magoo@example.com>\n" +
            r"250 Requested mail action okay, completed\n" +
            r"RCPT TO:<mrs.magoo@example.com>\n" +
            r"250 Requested mail action okay, completed\n" +
            r"DATA\n" +
            r"354 Start mail input; end with <CRLF>.<CRLF>\n" +
            r"(.*)\n\.\n" +
            r"250 Requested mail action okay, completed\n" +
            r"QUIT\n" +
            r"221 localhost Service closing channel$", log, re.S)
        self.failUnless(m)

        msg = rfc822.Message(StringIO.StringIO(m.group(2)))
        self.assertEqual(
            msg.getaddr("From"), ("Mr Magoo", "mr.magoo@example.com"))
        self.assertEqual(
            msg.getaddr("To"), ("Mrs Magoo", "mrs.magoo@example.com"))
        self.assertEqual(
            msg.getheader("Subject"), "Hello Mrs Magoo!")
        self.assertEqual(
            msg.fp.read(), "What's for tea Mrs Magoo?\n" +
            "I hope it's cherry pie cos I like that a lot\nLater!!!")

    # The XML syntax example JSP exercises the JSP XML document
    # functionality.
    def testXMLSyntax(self):
        r = self.getSession().GET("jsp/xml/xml.jsp")
        self.assertEqual(r.status, 200)
        self.assertEqual(r.type, "text/html")
        self.failUnless(re.search(
            r"^<html>.*" +
            r"This is the output of a simple JSP using XML format\..*" +
            r"1\n, \n2\n, \n3\n, \n4\n, \n5\n, \n.*" +
            r"\d\d?-\w\w\w-\d\d \d\d?:\d\d:\d\d.*" +
            r"<p>This sentence is enclosed in a jsp:text element\.</p>.*" +
            r"</html>$", r.body, re.S))
