# JDBC and DBCP test
# Coverage: Basic operations, and realms.
# Exercises: JNDI, JDBC, MySQL JDBC driver, PostgreSQL JDBC driver,
#   HTTP authentication, Role-based authentication, JDBC realms.

from harness.testsuite import TestCase

import MySQLdb
import pgdb
import sys
import types


##############################################################################
# Base class for all testcases
##############################################################################

class JDBCTestCase:
    deploy = 1

    # Test that basic JDBC operations work.
    def testBasicJDBC(self):
        s = self.getSession()
        s.setURL("test")

        # Check that we are using the database that we expect to be,
        # just in case there's some wierdness going on somewhere.
        module = sys.modules[self.__module__]
        for item in dir(module):
            item = getattr(module, item)
            if (type(item) == types.ClassType and
                issubclass(item, TestCase) and
                item.webapp is not None):
                assert item.webapp.endswith("-test")
                r = s.GET("?action=assert-" + item.webapp[:-5])
                self.assertEqual(r.status, 200)
                if item.webapp == self.webapp:
                    self.failUnless(r.body.startswith("OK"))
                else:
                    self.failUnless(r.body.startswith("FAIL"))

        # Create some tables using both individual and batched queries.
        self.clearTables("books", "authors", "authors_map")
        for action in ("create-single", "create-batch"):
            r = s.GET("?action=" + action)
            self.assertEqual(r.status, 200)
            self.failUnless(r.body.startswith("OK"))

        # Populate the tables we just created using prepared statements.
        tables = [
            ["books", [
                [1, "0-596-00283-1", "Java in a Nutshell"],
                [2, "0-596-00152-5", "Java Enterprise in a Nutshell"],
                [3, "0-349-10839-0", "Generation X"]]],
            ["authors", [
                [1, "David Flanagan"],
                [2, "Jim Farley"],
                [3, "William Crawford"],
                [4, "Douglas Coupland"]]],
            ["authors_map", [[1, 1], [2, 2], [2, 3], [2, 1], [3, 4]]]]
        r = s.POST("", "action=insert&data=" + "\n".join(
            ["\n".join([" ".join([t] + map(str, r)) for r in d])
             for t, d in tables]))
        self.assertEqual(r.status, 200)
        self.failUnless(r.body.startswith("OK"))
        self.checkTablesMatch(tables)

        # Check that selects, updates and deletes work.
        for action in (None, "update", "delete"):
            if action is not None:
                r = s.GET("?action=" + action)
                self.assertEqual(r.status, 200)
                self.failUnless(r.body.startswith("OK"))
                if action == "update":
                    tables[0][1][2][1] = "0-349-10331-3"
                elif action == "delete":
                    del tables[0][1][1]
            r = s.GET("?action=list")
            self.assertEqual(r.status, 200)
            self.assertEqual(r.lines, [
                "%s: %s: %s" % (isbn, title, ", ".join([
                    [name for auth_id, name in tables[1][1] if auth_id == a][0]
                    for b, a in tables[2][1] if b == book_id]))
                for book_id, isbn, title in tables[0][1]])


    # Check that JDBC realms using this database work as expected.
    def testJDBCRealm(self):
        expect = {"norole": 403, "wrongrole": 403, "privileged": 200}
        users = expect.keys()
        roles = [user for user in users if user != "norole"]

        # Populate the database
        self.clearTables("users", "roles")
        self.csr.execute("""\
            CREATE TABLE users (
              user_name CHAR(32) NOT NULL,
              user_pass CHAR(32) NOT NULL)""")
        for user in users:
            self.csr.execute("""\
                INSERT INTO users VALUES (%s, %s)""", (user, user))
        self.csr.execute("""\
            CREATE TABLE roles (
              user_name CHAR(32) NOT NULL,
              role_name CHAR(32) NOT NULL)""")
        for role in roles:
            self.csr.execute("""\
                INSERT INTO roles VALUES (%s, %s)""", (role, role))
        self.db.commit()

        # Peform the test
        s = self.getSession()
        r = s.GET("protected/file")
        self.assertEqual(r.status, 401)

        for user in users:
            for user, password, status in (
                (user + "x", user + "x", 401),
                (user + "x", user, 401),
                (user, user + "x", 401),
                (user, user, expect[user])):

                s.authenticate(user, password)
                r = s.GET("")
                self.assertEqual(r.status, status)
                if status == 200:
                    self.assertEqual(r.body, "Hello\n")

    # Utility stuff

    def setUp(self):
        try:
            self.db = self.driver.connect(
                host = "localhost.localdomain", db = "tomcattest",
                user = "tomcattest", passwd = "tomcattest")
        except:
            self.skip(sys.exc_info())
        self.csr = self.db.cursor()
        TestCase.setUp(self)

    def clearTables(self, *tables):
        for table in tables:
            try:
                self.csr.execute("DROP TABLE %s" % table)
            except:
                pass
            self.db.commit()

    def checkTablesMatch(self, tables):
        for name, data in tables:
            self.csr.execute("SELECT * from %s" % name)
            self.assertEqual(self.csr.rowcount, len(data))
            for row in data:
                self.assertEqual(list(self.csr.fetchone()), row)


##############################################################################
# Database abstraction mini-layer
##############################################################################

class Driver:
    convert_params = {}

    def connect(self, **params):
        for param in params.keys():
            newparam = self.convert_params.get(param)
            if newparam is not None:
                params[newparam] = params[param]
                del params[param]
        return apply(self.driver.connect, (), params)
        
class MySQLDriver(Driver):
    driver = MySQLdb
    
class PostgreSQLDriver(Driver):
    driver = pgdb
    convert_params = {"passwd": "password", "db": "database"}
    
        
##############################################################################
# The testcases themselves
##############################################################################

class MySQLTestCase(JDBCTestCase, TestCase):
    webapp = "mysql-test"
    driver = MySQLDriver()

class PostgreSQLTestCase(JDBCTestCase, TestCase):
    webapp = "pgsql-test"
    driver = PostgreSQLDriver()
