from harness.testsuite import TestCase

import os
import re
import zipfile
from harness import util

class LinkingTestCase(TestCase):

    # This test can be run anytime (it doesn't even need a running
    # server), but we run it right at the start, since a failure here
    # will probably cause failures all over the place.
    order = 0

    # Check that all Java shared libraries may be dlopened.  It is
    # possible that Tomcat will run normally with undlopenable
    # libraries, depending on the order in which they are loaded, but
    # it has a tendency to mask problems so should be avoided.
    def testDlopen(self):
        test = "%s %%s 2>&1" % os.path.join(self.getDataDir(), "dlopen")
        messages = ""
        for rpm in managedPackages():
            for item in packageContents(rpm):
                if item.endswith(".so"):
                    pipe = util.popen(test % item)
                    message = pipe.read()
                    if message:
                        self.failUnlessRaises(util.ExecError, pipe.close)
                    else:
                        pipe.close()
                    messages += message
        self.failIf(messages, "\n" + messages.rstrip())

    # Check that the the Java shared libraries and the jarfiles
    # contain the same set of resources.
    def testResources(self):
        libs = []
        jars = []
        for rpm in managedPackages():
            items = packageContents(rpm)
            thislibs = [item for item in items if item.endswith(".so")]
            thisjars = [item for item in items if item.endswith(".jar")]
            self.failIf(thislibs and not thisjars)
            if thislibs:
                libs.extend(thislibs)
                jars.extend(thisjars)

        jarresources = []
        for jar in jars:
            jarresources.extend(
                [item for item in zipfile.ZipFile(jar, "r").namelist()
                 if not (item.endswith(".class") or
                         item.endswith("/") or
                         item == "META-INF/MANIFEST.MF")])

        libresources = []
        pipe = util.popen(" ".join(
            ["/bin/sh", os.path.join(self.getDataDir(), "listresources.sh")] +
            libs))
        libresources = [line.rstrip() for line in pipe.xreadlines()]
        pipe.close()

        onlylibs = [r for r in libresources if r not in jarresources]
        onlyjars = [r for r in jarresources if r not in libresources]

        messages = ""
        if onlylibs:
            messages += "\n  ".join(
                ["The following resources exist only in shared libraries:"] +
                onlylibs) + "\n"
        if onlyjars:
            messages += "\n  ".join(
                ["The following resources exist only in jar files:"] +
                onlyjars) + "\n"

        self.failIf(messages, "\n" + messages.rstrip())

    # Check that the the Java shared libraries and the jarfiles
    # contain the same set of classes
    def testClasses(self):
        class_re = re.compile(r"^[0-9a-fA-F]*\s+D\s+(\S+)::class\$$")
        onlylibs = []
        onlyjars = []
        for rpm in managedPackages():
            items = packageContents(rpm)
            libs = [item for item in items if item.endswith(".so")]
            jars = [item for item in items if item.endswith(".jar")]
            if libs:
                self.failIf(not jars)
            else:
                continue

            jarclasses = []
            for jar in jars:
                jarclasses.extend(
                    [item[:-6].replace("/", ".")
                     for item in zipfile.ZipFile(jar, "r").namelist()
                     if item.endswith(".class")])

            libclasses = []
            pipe = util.popen(" ".join(["/usr/bin/nm", "-DC"] + libs))
            for line in pipe.xreadlines():
                m = class_re.search(line.rstrip())
                if m:
                    libclasses.append(m.group(1).replace("::", "."))
            pipe.close()

            onlylibs.extend([r for r in libclasses if r not in jarclasses])
            onlyjars.extend([r for r in jarclasses if r not in libclasses])

        messages = ""
        if onlylibs:
            messages += "\n  ".join(
                ["The following classes exist only in shared libraries:"] +
                onlylibs) + "\n"
        if onlyjars:
            messages += "\n  ".join(
                ["The following classes exist only in jar files:"] +
                onlyjars) + "\n"

        self.failIf(messages, "\n" + messages.rstrip())

# Utility stuff

def managedPackages():
    """Return a list of all Java rpms managed by redhat-java-rpm-scripts."""
    pipe = util.popen("/bin/rpm -q --whatrequires redhat-java-rpm-scripts")
    try:
        return [line.rstrip() for line in pipe.xreadlines()]
    finally:
        pipe.close()

def packageContents(package):
    """Return a list of all files owned by the specified rpm."""
    pipe = util.popen("/bin/rpm -ql %s" % package)
    try:
        return [line.rstrip() for line in pipe.xreadlines()]
    finally:
        pipe.close()
