# Basic load test
# Coverage: Fairly simplistic.
# Exercises: JSPs, Performance under load.

from harness.testsuite import TestCase
from harness import log

import os
import re
import sys
import thread
import time
import traceback

# Parameters for the test
NUM_WORKERS = 50
TEST_LENGTH = 60

# States for the state machine
STATE_WAIT = 1
STATE_RUN  = 2
STATE_EXIT = 3

class LoadTestCase(TestCase):
    webapp = "load-test"
    deploy = 1

    # Check that the server can take a battering.
    def testLoad(self):
        s = self.getSession()

        # Ensure that requests to the root of the webapp are
        # redirected to the index JSP.
        r = s.GET("")
        self.assertEqual(r.status, 302)
        self.failUnless(r.headers["Location"].endswith(
            "/" + self.webapp + "/index.jsp"))

        # Ensure that the index JSP is being executed, and pull the
        # list of images out of it.  This has the side effect of
        # compiling the JSP if it wasn't already, so that the workers
        # don't end up wasting valuable test time waiting for that.
        r = s.GET("index.jsp")
        self.assertEqual(r.status, 200)
        self.failUnless(re.search(
            r"<li>%(os-name)s\s+<li>%(os-version)s\s+<li>%(os-architecture)s" %
            self.server.manager.serverinfo(), r.body))
        self.urls = [("", 302, None), ("index.jsp", 200, r.body)] + [
            (f, 200, open(os.path.join(
                self.getDataDir(), self.webapp, f), "r").read())
            for f in self.getImages(r.body)]

        # Spawn the workers and run the test
        self.mutex = thread.allocate_lock()
        self.count = 0
        self.state = STATE_WAIT

        self.requests = 0
        self.exceptions = []

        for i in range(NUM_WORKERS):
            thread.start_new_thread(self.worker, ())
        while self.count < NUM_WORKERS:
            pass
        self.state = STATE_RUN
        time.sleep(TEST_LENGTH)
        self.state = STATE_EXIT
        while self.count > 0:
            pass

        # Print the results and check that everything went OK.
        log.debug("%d requests completed during load test." % self.requests)
        self.failIf(self.exceptions, "The following exceptions occurred:\n\n" +
                    "".join(["".join(apply(traceback.format_exception, e))
                             for e in self.exceptions]))

    def worker(self):
        s = self.getSession()
        baseurl = s.resolveURL("")

        self.mutex.acquire()
        self.count += 1
        self.mutex.release()
        while self.state == STATE_WAIT:
            pass

        num_requests = 0
        exception = None
        try:
            while 1:
                for url, status, body in self.urls:
                    if self.state != STATE_RUN:
                        break
                    r = s.GET(baseurl + url)
                    self.assertEqual(r.status, status)
                    if body is not None:
                        self.assertEqual(r.body, body)
                    num_requests += 1
                else:
                    continue
                break
        except:
            exception = sys.exc_info()

        self.mutex.acquire()
        log.debug("load test worker %d completed %d requests." % (
            thread.get_ident(), num_requests))
        self.requests += num_requests
        if exception is not None:
            self.exceptions.append(exception)
        self.count -= 1
        self.mutex.release()

    # Utility stuff

    def getImages(self, html):
        list = []
        search_re = re.compile(r'^.*?<img\s+src="([^"]+)"(.*)$', re.S)
        while 1:
            m = search_re.search(html)
            if not m:
                break
            file, html = m.groups()
            list.append(file)
        return list
