# Tomcat Manager Application test
# Coverage: Everything but the deploy and undeploy commands.
# Exercises: HTTP authentication, Role-based authentication,
#   File-based realms.

from harness.testsuite import TestCase

import copy
import os
import re
from harness import util

class ManagerTestCase(TestCase):
    webapp = "manager"

    # Check that authentication works as expected
    def testAuthentication(self):
        for url, type in (("html", "text/html"), ("list", "text/plain")):
            s = self.getSession()
            r = s.GET(url)
            self.assertEqual(r.status, 401)

            for (user, password, status) in (
                (self.server.manageruser + "x", self.server.managerpass, 401),
                (self.server.manageruser, self.server.managerpass + "x", 401),
                (self.server.adminuser, self.server.adminpass, 403),
                (self.server.manageruser, self.server.managerpass, 200)):

                s.authenticate(user, password)
                r = s.GET("")
                self.assertEqual(r.status, status)
                if status == 200:
                    self.assertEqual(r.type, type)
                    self.failUnless(self.getMessage(r).startswith("OK"))

    # Check that we cannot reload, remove or stop ourself
    def testNoSelfManagement(self):
        expected = \
          "FAIL - The manager can not reload, remove, stop, or undeploy itself"
        for html in (0, 1):
            s = self.getSession()
            if html:
                s.setURL("html/")
            s.authenticate(self.server.manageruser, self.server.managerpass)
            for action in ("reload", "remove", "stop"):
                r = s.GET(action + "?path=/" + self.webapp)
                self.assertEqual(self.getMessage(r), expected)

    # Check that we can install, start, stop, reload and remove
    # webapps as directories, config.xml files or wars.
    def testWebappManagement(self):
        testapp = "manager-test"
        ts = self.getSession(testapp)

        datadir = self.getDataDir()
        appsdir = self.getAppsDir()
        apps = self.listApps()

        for html in (0, 1):
            s = self.getSession()
            if html:
                s.setURL("html/")
            s.authenticate(self.server.manageruser, self.server.managerpass)
            for type in ("dir", "war", "xml"):
                for inappbase in (1, 0):
                    path = testapp
                    if type != "dir":
                        path += ".%s" % type
                    if inappbase:
                        util.system("/bin/cp", "-a",
                                    os.path.join(datadir, path), appsdir)
                        if type != "xml":
                            source = path
                        path = os.path.join(appsdir, path)
                        if type == "xml":
                            source = "file://" + path
                    else:
                        path = os.path.join(datadir, path)
                        source = "file://" + path

                    for action in ("install","stop","start","reload","remove"):
                        # Perform the action
                        url = action + "?"
                        if action == "install":
                            what = {"xml": "config"}.get(type, "war")
                            if html:
                                what = "install" + what.title()
                            url += "%s=%s" % (what, source)
                        else:
                            url += "path=/%s" % testapp
                        r = s.GET(url)

                        # Check that the action was reported done as expected
                        self.assertEqual(r.status, 200)
                        if (action, type, inappbase) == ("reload", "war", 0):
                            expect = "FAIL - Reload not supported " + \
                                     "on WAR deployed at path /" + testapp
                        else:
                            expect = "OK - %sed application " % \
                                     {"stop": "stopp", "remove": "remov"}.get(
                                         action, action).title()
                            if (action, type) == ("install", "xml"):
                                expect += "from context file " + source
                            else:
                                expect += "at context path /" + testapp
                        self.assertEqual(self.getMessage(r), expect)

                        # Check that the action actually happened as expected
                        old = copy.copy(apps)
                        new = []
                        for app in self.listApps():
                            if app in old:
                                del old[old.index(app)]
                            else:
                                new.append(app)

                        running = action not in ("stop", "remove")
                        self.assertEqual(old, [])
                        if action == "remove":
                            self.assertEqual(new, [])
                        else:
                            expect = os.path.realpath(path)
                            if type == "xml":
                                expect = os.path.join(datadir, testapp)
                            elif inappbase:
                                if type == "dir":
                                    expect += "/"
                                elif type == "war":
                                    expect = expect[:-4]
                            self.assertEqual(new, [[
                                "/" + testapp,
                                running and "running" or "stopped",
                                "0", expect]])

                        # Check that the app is working properly
                        if running:
                            for item in ("static", "servlet"):
                                r = ts.GET(item)
                                self.assertEqual(r.status, 200)
                                self.assertEqual(
                                    r.body,
                                    "Hello world! (%s)\r\n" % item)

                    if inappbase:
                        if type == "xml":
                            util.system("/bin/rm", "-f", path)
                        else:
                            self.failIf(os.path.exists(path))

    # Check that the writable webapps detection thing works
    def testWritableDirCheck(self):
        s = self.getSession()
        s.authenticate(self.server.manageruser, self.server.managerpass)

        dir = os.path.realpath(self.getAppsDir())
        writable_re = re.compile("Select WAR file to upload")
        unwritable_re = re.compile("<tt>%s</tt> unwritable" % dir)

        oldmode = os.stat(dir).st_mode
        os.chmod(dir, 0555)
        try:
            r = s.GET("html")
            self.assertEqual(r.status, 200)
            self.failUnless(unwritable_re.search(r.body))
            self.failIf(writable_re.search(r.body))
        finally:
            os.chmod(dir, oldmode)
            
        r = s.GET("")
        self.assertEqual(r.status, 200)
        self.failUnless(writable_re.search(r.body))
        self.failIf(unwritable_re.search(r.body))

    # XXX Check that we can deploy and undeploy .war files

    # Utility stuff

    html_msg_re = re.compile(
        '<td class="row-left" width="10%">' +
        '<small><b>Message:</b></small>&nbsp;</td>\s*' +
        '<td class="row-left"><pre>([^<]+)</pre></td>')

    def getMessage(self, response):
        if response.type == "text/html":
            m = self.html_msg_re.search(response.body)
            self.failUnless(m)
            text = m.group(1).rstrip()
        elif response.type == "text/plain":
            text = response.lines[0]
        else:
            raise ValueError, "bad content-type '%s'" % response.type
        return text
        
    def listApps(self):
        s = self.getSession()
        s.authenticate(self.server.manageruser, self.server.managerpass)
        r = s.GET("list")
        self.assertEqual(r.type, "text/plain")
        self.failUnless(r.lines[0].startswith("OK - Listed applications"))
        apps = [line.split(":") for line in r.lines[1:]]
        apps.sort()
        return apps
